Unit	UConsole;


//	"Open_Console" creates a console window for the application. It sets "no_console" to
//	True or False, depending on success. "Close_Console" closes the console window. It is
//	automatically called when the unit reaches the "Finalization" section.
//
//	"Set_TextAttr" changes the text and background color for following writes. "ClrScr"
//	clears the entire window. "GotoXY", "Set_CursorSize" and "ShowCursor" can be used for
//	setting the cursor position and visibility.


{$ASSERTIONS ON}						// assertions MUST be enabled


interface  ////////////////////////////////////////////////////////////////////////////////////


type	TConsoleColor	=	0..  15			;	// just like in the DOS days
	TConsoleCoord	=	1..4096			;	// should be sufficient
	TCursorSize	=	1.. 100			;	// 100 = full


const	ccBlack		:	TConsoleColor	=   0	;	// console color constants
	ccBlue		:	TConsoleColor	=   1	;
	ccGreen		:	TConsoleColor	=   2	;
	ccCyan		:	TConsoleColor	=   3	;
	ccRed		:	TConsoleColor	=   4	;
	ccMagenta	:	TConsoleColor	=   5	;
	ccBrown		:	TConsoleColor	=   6	;
	ccLightGray	:	TConsoleColor	=   7	;
	ccDarkGray	:	TConsoleColor	=   8	;
	ccLightBlue	:	TConsoleColor	=   9	;
	ccLightGreen	:	TConsoleColor	=  10	;
	ccLightCyan	:	TConsoleColor	=  11	;
	ccLightRed	:	TConsoleColor	=  12	;
	ccLightMagenta	:	TConsoleColor	=  13	;
	ccYellow	:	TConsoleColor	=  14	;
	ccWhite		:	TConsoleColor	=  15	;


var	ConsoleColumn	:	String		= '    ';
	ConsolePrefix	:	String		= '- '	;	// default is for message logging
	no_console	:	Boolean		= True	;


function  Close_Console		(				) : Boolean	;
procedure ClrScr		(				)		;
procedure Console_SHL		(				)		;
procedure Console_SHR		(				)		;
function  Get_CursorSize	(				) : TCursorSize	;
procedure GotoXY		(const	x, y  : TConsoleCoord	)		;
function  Open_Console		(const	Title : String		) : Boolean	;
procedure Set_CursorSize	(const  Value : TCursorSize	)		;
procedure Set_TextBackground	(const	Color : TConsoleColor	)		;
procedure Set_TextColor		(const	Color : TConsoleColor	)		;
procedure ShowCursor		(const	Value : Boolean		)		;
function  WhereX		(				) : TConsoleCoord;
function  WhereY		(				) : TConsoleCoord;
procedure Write			(				)		;  overload;
procedure Write			(const	Text  : String		)		;  overload;
procedure WriteLn		(				)		;  overload;
procedure WriteLn		(const	Text  : String		)		;  overload;
procedure WriteTest		(				)		;


implementation  ///////////////////////////////////////////////////////////////////////////////
uses	Windows, SysUtils;


const	Bit0		=	DWord(1) SHL 0;
	Bit1		=	DWord(1) SHL 1;
	Bit2		=	DWord(1) SHL 2;
	Bit3		=	DWord(1) SHL 3;

	CRLF		=	#13#10;

	DefaultColor	=	7 + 0;				// light gray


var	ConError	:	THandle;
	ConInput	:	THandle;
	ConOutput	:	THandle;

	ConsoleHWND	:	HWND;
	cur_Color	:	Word;
	xRes		:	Word;
	yRes		:	Word;


// tools


function ConsoleCtrlHandler(CtrlType : DWord) : Boolean;  stdcall;
const	Mask		=	[CTRL_C_EVENT, CTRL_BREAK_EVENT, CTRL_CLOSE_EVENT];
begin
Result := (CtrlType in Mask);					// True = no further processing
end;


function Get_ConsoleWindow : HWND;
var	tmpStr		:	String;
begin
tmpStr := IntToStr(GetTickCount) + ' ' + IntToStr(GetCurrentProcessID);
Assert(SetConsoleTitle(pChar(tmpStr)));
Sleep(40);
Result := FindWindow(NIL, pChar(tmpStr));
Assert(Result <> 0);
end;


procedure Get_Handle(var Handle : tHandle);
begin
if (@Handle = @ConError ) then Handle := GetStdHandle(STD_ERROR_HANDLE ) else
if (@Handle = @ConInput ) then Handle := GetStdHandle(STD_INPUT_HANDLE ) else
if (@Handle = @ConOutput) then Handle := GetStdHandle(STD_OUTPUT_HANDLE) else Assert(False);
Assert(Handle <> INVALID_HANDLE_VALUE);
end;


procedure Reset_Colors;
begin
Set_TextColor(ccLightGray);					// DOS default
Set_TextBackground(ccBlack);
end;


// public functions


procedure Console_SHL;						// less indentation
begin
Delete(ConsolePrefix, 1, Length(ConsoleColumn));
end;


procedure Console_SHR;						// more indentation
begin
ConsolePrefix := ConsoleColumn + ConsolePrefix;
end;


procedure ClrScr;						// clear screen
var	CharsWritten	:	DWord;
	Position	:	tCoord;
begin
if no_console then Exit;
Assert(FillConsoleOutputCharacter(ConOutput, ' ', xRes * yRes, Position, CharsWritten));
GotoXY(1, 1);
end;


function Close_Console : Boolean;
begin
if no_console then begin
	Result := False;
	Exit;
end;
Result := FreeConsole;
no_console := Result;
end;


function Get_CursorSize : TCursorSize;
var	CursorInfo	:	tConsoleCursorInfo;
begin
if no_console then begin
	Result := 100;
	Exit;
end;
Assert(GetConsoleCursorInfo(ConOutput, CursorInfo));
Result := CursorInfo.dwSize;
end;


procedure GotoXY(const x, y : TConsoleCoord);
var	Position	:	tCoord;
begin
if no_console then Exit;
Position.x := x - 1;
Position.y := y - 1;
Assert(SetConsoleCursorPosition(ConOutput, Position));
end;


function Open_Console(const Title : String) : Boolean;
const	Flags		=	SWP_NOMOVE OR SWP_NOSIZE;
var	Position	:	tCoord;
begin
if no_console then begin					// create new console window
	Result := AllocConsole;
	no_console := (not Result);
	if no_console then Exit;
	cur_Color   := DefaultColor;
	ConsoleHWND := Get_ConsoleWindow;			// get window handle
	Assert(SetConsoleTitle(pChar(Title)));			// set title
	Assert(SetConsoleOutputCP(437));			// set codepage (needs TTF font)
	Get_Handle(ConError );					// get standard file handles
	Get_Handle(ConInput );
	Get_Handle(ConOutput);
	Assert(SetWindowPos(ConsoleHWND, HWND_TOPMOST, 0, 0, 0, 0, Flags));		// topmost
	Assert(RemoveMenu(GetSystemMenu(ConsoleHWND, False), SC_CLOSE, MF_BYCOMMAND));	// no [X]
	Assert(SetConsoleCtrlHandler(@ConsoleCtrlHandler, True));			// no events
	Position := GetLargestConsoleWindowSize(ConOutput);				// get size
	with Position do begin
		Assert((x <> 0) and (y <> 0));
		xRes := x;
		yRes := y;
	end;
end else begin							// console already exists, reset
	Result := True;
	Assert(SetConsoleTitle(pChar(Title)));
	ShowCursor(True);
end;
Reset_Colors;
ClrScr;
end;


procedure Set_CursorSize(const Value : TCursorSize);
var	CursorInfo	:	tConsoleCursorInfo;
begin
if no_console then Exit;
Assert(GetConsoleCursorInfo(ConOutput, CursorInfo));  CursorInfo.dwSize := Value;
Assert(SetConsoleCursorInfo(ConOutput, CursorInfo));
end;


procedure Set_TextBackground(const Color : TConsoleColor);
const	all_bits	=	BACKGROUND_RED
			OR	BACKGROUND_GREEN
			OR	BACKGROUND_BLUE
			OR	BACKGROUND_INTENSITY;
begin
if no_console then Exit;
cur_Color := cur_Color OR all_bits;						// set all bits
if (Color AND Bit0 = 0) then cur_Color := cur_Color XOR BACKGROUND_BLUE;	// turn bits off
if (Color AND Bit1 = 0) then cur_Color := cur_Color XOR BACKGROUND_GREEN;
if (Color AND Bit2 = 0) then cur_Color := cur_Color XOR BACKGROUND_RED;
if (Color AND Bit3 = 0) then cur_Color := cur_Color XOR BACKGROUND_INTENSITY;
Assert(SetConsoleTextAttribute(ConOutput, cur_Color));				// set cur. col.
end;


procedure Set_TextColor(const Color : TConsoleColor);
const	all_bits	=	FOREGROUND_RED
			OR	FOREGROUND_GREEN
			OR	FOREGROUND_BLUE
			OR	FOREGROUND_INTENSITY;
begin
if no_console then Exit;
cur_Color := cur_Color OR all_bits;						// set all bits
if (Color AND Bit0 = 0) then cur_Color := cur_Color XOR FOREGROUND_BLUE;	// turn bits off
if (Color AND Bit1 = 0) then cur_Color := cur_Color XOR FOREGROUND_GREEN;
if (Color AND Bit2 = 0) then cur_Color := cur_Color XOR FOREGROUND_RED;
if (Color AND Bit3 = 0) then cur_Color := cur_Color XOR FOREGROUND_INTENSITY;
Assert(SetConsoleTextAttribute(ConOutput, cur_Color));				// set cur. col.
end;


procedure ShowCursor(const Value : Boolean);
var	CursorInfo	:	tConsoleCursorInfo;
begin
if no_console then Exit;
Assert(GetConsoleCursorInfo(ConOutput, CursorInfo));  CursorInfo.bVisible := Value;
Assert(SetConsoleCursorInfo(ConOutput, CursorInfo));
end;


function WhereX : TConsoleCoord;
var	Info		:	_CONSOLE_SCREEN_BUFFER_INFO;
begin
Assert(GetConsoleScreenBufferInfo(ConOutput, Info));
Result := Info.dwCursorPosition.x + 1;
end;


function WhereY : TConsoleCoord;
var	Info		:	_CONSOLE_SCREEN_BUFFER_INFO;
begin
Assert(GetConsoleScreenBufferInfo(ConOutput, Info));
Result := Info.dwCursorPosition.y + 1;
end;


procedure Write;
begin
if no_console then Exit;
Write(' ');
end;


procedure Write(const Text : String);
var	CharsWritten	:	LongWord;
begin
if no_console then Exit;
Assert(WriteConsole(ConOutput, pChar(Text), Length(Text), CharsWritten, NIL));
end;


procedure WriteLn;
begin
if no_console then Exit;
Write(CRLF);
end;


procedure WriteLn(const Text : String);
begin
if no_console then Exit;
Write(ConsolePrefix + Text + CRLF);
end;


procedure WriteTest;
const	List		=	[7, 8, 9, 10, 13];
var	i		:	Word;
	x		:	Word;
	y		:	Word;
begin
WriteLn('console unit write test');
Console_SHR;
i := 0;
for y := 0 to 15 do begin
	Set_TextBackground(0);  Write(ConsolePrefix);
	Set_TextBackground(y);
	for x := 0 to 15 do begin
		Set_TextColor(x);
		if (i in List)	then Write(' ')
				else Write(Chr(i));
		Inc(i);
	end;
	WriteLn;
end;
Console_SHL;
Reset_Colors;
end;


///////////////////////////////////////////////////////////////////////////////////////////////


initialization


finalization
Close_Console;


end.
