unit	UTileDecoders;


{
	The SNES uses small 'tiles' of 8x8 pixels to build its graphics. These tiles can be in
	various color depths. The PPU chips use tiles of 2, 4 and 8 bits per pixel. ROMs also
	use tiles of 1 and 3 bits per pixel to save space - at runtime these can be converted to
	useable ones simply by adding some bytes to the end.


	1 bpp  =   8 bytes per tile
	2 bpp  =  16 bytes per tile
	3 bpp  =  24 bytes per tile
	4 bpp  =  32 bytes per tile
	8 bpp  =  64 bytes per tile


	Tiles with 1 bit per pixel are the easiest: One row (8 pixels) is stored in one byte.
	For some reason the pixels are reversed though:

	column:	0	1	2	3	4	5	6	7	(left to right)
	----------------------------------------------------------------------------------------
	row 0:	bit 7	bit 6	bit 5	bit 4	bit 3	bit 2	bit 1	bit 0	=>  byte 0
	row 1:	bit 7	bit 6	bit 5	bit 4	bit 3	bit 2	bit 1	bit 0	=>  byte 1
	row 2:	bit 7	bit 6	bit 5	bit 4	bit 3	bit 2	bit 1	bit 0	=>  byte 2
	row 3:	bit 7	bit 6	bit 5	bit 4	bit 3	bit 2	bit 1	bit 0	=>  byte 3
	row 4:	bit 7	bit 6	bit 5	bit 4	bit 3	bit 2	bit 1	bit 0	=>  byte 4
	row 5:	bit 7	bit 6	bit 5	bit 4	bit 3	bit 2	bit 1	bit 0	=>  byte 5
	row 6:	bit 7	bit 6	bit 5	bit 4	bit 3	bit 2	bit 1	bit 0	=>  byte 6
	row 7:	bit 7	bit 6	bit 5	bit 4	bit 3	bit 2	bit 1	bit 0	=>  byte 7


	Tiles with 2 bits per pixel are stored similarly, but each row is stored in two
	bytes: the first byte for bit 0 of each pixel, and the second one for bit 1. This
	is Nintendo's basic unit for colored (NES, SNES) or shaded (GB) bitplane graphics.

	byte 0					byte 1
	-------------------------------------------------------------------
	bit 7	=  bit 0 of pixel 0		bit 7	=  bit 1 of pixel 0
	bit 6	=  bit 0 of pixel 1		bit 6	=  bit 1 of pixel 1
	bit 5	=  bit 0 of pixel 2		bit 5	=  bit 1 of pixel 2
	bit 4	=  bit 0 of pixel 3		bit 4	=  bit 1 of pixel 3
	bit 3	=  bit 0 of pixel 4		bit 3	=  bit 1 of pixel 4
	bit 2	=  bit 0 of pixel 5		bit 2	=  bit 1 of pixel 5
	bit 1	=  bit 0 of pixel 6		bit 1	=  bit 1 of pixel 6
	bit 0	=  bit 0 of pixel 7		bit 0	=  bit 1 of pixel 7


	Some ROMs use tiles with 3 bits per pixel (8 colors). Bits 0 and 1 of each pixel are
	stored in a regular 2-bpp tile structure in the first 16 bytes, and bit 2 is in a 1-bpp
	tile structure in the remaining 8 bytes.

	Tiles with 4 bits per pixel are stored in two 2-bpp structures. Bit 0 and 1 are in the
	first structure, bits 2 and 3 in the second.

	Tiles with 8 bits per pixel are stored in four 2-bpp structures, similarly to the 4-bpp
	tile. They are used by the SNES for background tiles in BG mode 3 and 4.

	The bits of Mode7 tiles are not split into bitplanes. These 256-color pixels have their
	palette index stored in a byte, just like on a PC. The only difference is that in VRAM,
	the tilemap bytes and the tile bytes are interleaved with each other.
}


interface


type	TTileLine	=	array[0..7] of Byte;			//  8 unpacked pixels
	TTile		=	array[0..7] of TTileLine;		// 64 unpacked pixels
	TLineOfTiles	=	array[Word] of TTileLine;		// tile bitmap scanline


	TBGTile		=	array[0..15, 0..31] of Byte;		// 32  16 pixels
	TSpriteTile	=	array[0..63, 0..63] of Byte;		// 64  64 pixels


	TTileDecoder	=	procedure(var PackedTile, DestTile);	// function pointer


	TPackedRow_1bpp	=	packed array[0..0] of Byte;		// 8 bits
	TPackedRow_2bpp	=	packed array[0..1] of Byte;		// 8 bits  2

	TBitplane_1bpp	=	packed array[0..7] of TPackedRow_1bpp;
	TBitplane_2bpp	=	packed array[0..7] of tPackedRow_2bpp;	// basic tile unit


	TPackedTile1	=	packed record				// 1 bpp
				Plane_0		:	TBitplane_1bpp;
				end;


	TPackedTile2	=	packed record				// 2 bpp
				Plane_0		:	TBitplane_2bpp;
				end;


	TPackedTile3	=	packed record				// 2 bpp + 1 bpp
				Plane_0		:	TBitplane_2bpp;
				Plane_1		:	TBitplane_1bpp;
				end;


	TPackedTile4	=	packed record				// 2 bpp  2
				Plane_0		:	TBitplane_2bpp;
				Plane_1		:	TBitplane_2bpp;
				end;


	TPackedTile8	=	packed record				// 2 bpp  4
				Plane_0		:	TBitplane_2bpp;
				Plane_1		:	TBitplane_2bpp;
				Plane_2		:	TBitplane_2bpp;
				Plane_3		:	TBitplane_2bpp;
				end;


	TPackedPixel7i	=	packed array[0..1] of Byte;		//     interleaved pixel
	TPackedPixel7n	=	packed array[0..0] of Byte;		// non-interleaved pixel

	tPackedRow7i	=	packed array[0..7] of tPackedPixel7i;	//     interleaved pixels
	tPackedRow7n	=	packed array[0..7] of tPackedPixel7n;	// non-interleaved pixels

	tPackedTile7i	=	packed array[0..7] of tPackedRow7i;	//     interleaved tile
	tPackedTile7n	=	packed array[0..7] of tPackedRow7n;	// non-interleaved tile


procedure TileDecoder1		(var PackedTile, DestTile);		// forward declarations
procedure TileDecoder2		(var PackedTile, DestTile);
procedure TileDecoder3		(var PackedTile, DestTile);
procedure TileDecoder4		(var PackedTile, DestTile);
procedure TileDecoder7I		(var PackedTile, DestTile);
procedure TileDecoder7N		(var PackedTile, DestTile);
procedure TileDecoder8		(var PackedTile, DestTile);


implementation
uses	UShared;


{$INCLUDE Tiledecoders\output\1 }	// If Turbo Delphi reports that it can't find...
{$INCLUDE Tiledecoders\output\2 }	// one of these files, try restarting Turbo Delphi.
{$INCLUDE Tiledecoders\output\3 }
{$INCLUDE Tiledecoders\output\4 }
{$INCLUDE Tiledecoders\output\7i}
{$INCLUDE Tiledecoders\output\7n}
{$INCLUDE Tiledecoders\output\8 }


end.
